"""
Check the existence and, optionally, the version of available executables or
commands.
"""
import typing as t
from shutil import which
import subprocess

from .version import CheckVersion
from .utils import version_to_tuple
from ..config import Parameter

__all__ = ("CheckExec",)


class CheckExec(CheckVersion):
    """Check for the presence and version of executables"""

    #: The executable may exist and be installed, but get_current_version
    #: may not be able to identify the current version
    require_current_version = False

    msg = Parameter(
        "CHECK_EXEC.MSG",
        default="Check executable '{check.raw_value}'",
    )

    aliases = ("checkExec",)

    @property
    def value(
        self,
    ) -> t.Tuple[
        t.Union[str, None], t.Union[t.Callable, None], t.Union[t.Tuple[int], None]
    ]:
        """Get the package name, comparison operator and version tuple."""
        cmd_name, op, version = CheckVersion.value.fget(self)

        # Check to see if the cmd_name exists. Returns None if it doesn't
        cmd_name = which(cmd_name) if cmd_name is not None else cmd_name

        return cmd_name, op, version

    @value.setter
    def value(self, v):
        CheckVersion.value.fset(self, v)

    def get_current_version(self) -> t.Union[None, t.Tuple[int]]:
        """Try to get the version tuple for the executable."""
        cmd_name, op, version = self.value

        if cmd_name is None:  # command not found
            return None

        for args in (  # Different commands to try for versions
            (cmd_name, "-V"),
            (cmd_name, "--version"),
        ):
            try:
                proc = subprocess.run(args, capture_output=True)
            except FileNotFoundError:
                # Couldn't find the executable
                continue

            if proc.returncode != 0:  # Wasn't a success
                continue

            # Try to parse the current version string
            current_version = version_to_tuple(proc.stdout.decode("UTF-8"))
            current_version = (
                current_version
                if current_version is not None
                else version_to_tuple(proc.stderr.decode("UTF-8"))
            )

            if current_version is not None:
                # Current version found! We're done
                return current_version

        # Not found
        return None



"""Utility functions"""
import typing as t
import operator
import re

__all__ = ("all_subclasses", "pop_first", "version_to_tuple", "name_and_version")

__missing__ = object()  # used an argument for missing values


def all_subclasses(cls) -> t.List[t.Type]:
    """Retrieve all subclasses, sub-subclasses and so on for a class

    Parameters
    ----------
    cls : Type
        The class object to inspect for subclasses.

    Returns
    -------
    subclasses : list
        The list of all subclasses.

    Examples
    --------
    >>> class A(object): pass
    >>> class B(A): pass
    >>> class C(B): pass
    >>> all_subclasses(A)
    [<class 'geomancy.checks.utils.B'>, <class 'geomancy.checks.utils.C'>]
    """
    return cls.__subclasses__() + [
        g for s in cls.__subclasses__() for g in all_subclasses(s)
    ]


def pop_first(d: dict, *keys, del_remaining: bool = True, default: t.Any = __missing__):
    """Pop the first key found in the dict.

    Parameters
    ----------
    d
        The dict to search for keys to pop
    *keys
        The list of keys to search
    del_remaining
        Delete the remaining keys from the dict
    default
        Return this value if a key isn't found. If not specified, a KeyError
        is raised

    Returns
    -------
    value
        The value for the first matching key

    Raises
    ------
    KeyError
        If none of the keys were found and a default wasn't found

    Examples
    --------
    >>> pop_first({'a': 1, 'b': 2, 'c': 3}, 'b', 'a')
    2
    >>> pop_first({'a': 1, 'b': 2, 'c': 3}, 'a', 'b')
    1
    >>> pop_first({'a': 1, 'b': 2, 'c': 3}, 'e', default='missing')
    'missing'
    >>> d = {'a1': 1, 'a2': 2, 'b': 3}
    >>> pop_first(d, 'a1', 'a2')
    1
    >>> print(d)
    {'b': 3}
    >>> pop_first(d, 'a1')
    Traceback (most recent call last):
        ...
    KeyError: "Could not find any of the keys: ('a1',)"
    """
    value = __missing__

    # Find the first key
    for key in keys:
        value = d.pop(key, __missing__)
        if value is not __missing__:
            break

    # Del remaining
    if del_remaining:
        extra_keys = [key for key in keys if key in d]
        for k in extra_keys:
            del d[k]

    # Return a value or raise an exception
    if value is not __missing__:
        return value
    elif default is not __missing__:
        return default
    else:
        raise KeyError(f"Could not find any of the keys: {keys}")


def version_to_tuple(version: str) -> t.Union[t.Tuple[int], None]:
    """Convert a semantic version string to a tuple of integers.

    Parameters
    ----------
    version
        String with the version number. e.g. '1.4.3]

    Returns
    -------
    version_tuple or None
        Vesion in the form of a tuple of integers, or
        None if a version couldn't be found

    Notes
    -----
    This function ignores letters in a version. e.g. '0.4.1b' becomes (0, 4, 1)

    Examples
    --------
    >>> version_to_tuple('26')
    (26,)
    >>> version_to_tuple('1.3')
    (1, 3)
    >>> version_to_tuple('v0.1.3')
    (0, 1, 3)
    >>> version_to_tuple('1.2beta')
    (1, 2)
    """
    version_match = re.search(r"([\d\.]{2,}|\d+)", version)  # Capture the version
    if version_match is None:
        return None
    version = version_match.group().split(".")  # Split the version string at "."
    return tuple(map(int, version))  # Convert to a tuple of ints


def name_and_version(
    string,
) -> t.Tuple[
    t.Union[None, str], t.Union[None, t.Callable], t.Union[None, t.Tuple[int]]
]:
    """Parse the name and version from a package name string with an optional
    specifier.

    Parameters
    ----------
    string
        The string with a name and version to parase.

    Returns
    -------
    name, operator, version
        The name, operator and version for a package.
        None, if the string couldn't be parsed.

    Examples
    --------
    >>> name_and_version('test')
    ('test', None, None)
    >>> name_and_version('test-ab_one')
    ('test-ab_one', None, None)
    >>> name_and_version('test-ab_one>=0.3.1')
    ('test-ab_one', <built-in function ge>, (0, 3, 1))
    >>> name_and_version('new-package>1.2')
    ('new-package', <built-in function gt>, (1, 2))
    >>> name_and_version('jinja2>=2.10')
    ('jinja2', <built-in function ge>, (2, 10))
    >>> name_and_version('geo>=0.4.0a')
    ('geo', <built-in function ge>, (0, 4, 0))
    """

    m = re.search(
        r"([\w\.\-_]+)"  # match package name
        r"\s*(==|>=|<=|>|<|!=)?\s*"  # operator
        r"([\d\w\.]+)?",  # version number
        string,
    )

    if m is None:
        # The string couldn't be parsed
        return None, None, None

    name = m.group(1)
    op = m.group(2)
    version = m.group(3)

    # Convert the operator
    if op is not None:
        op_mapping = {
            "==": operator.eq,
            ">": operator.gt,
            "<": operator.lt,
            ">=": operator.ge,
            "<=": operator.le,
            "!=": operator.ne,
        }
        op = op_mapping[op]

    # Convert the version to a tuple of integers
    version = version_to_tuple(version) if version is not None else None

    return name, op, version



"""Check abstract base class for checks that include versions"""
import typing as t
from abc import abstractmethod

from .base import Check, Result, Executor
from .utils import name_and_version

__all__ = ("CheckVersion",)


class CheckVersion(Check):
    """An abstract Check for package and program versions"""

    #: If true, the result of get_current_version must not be None
    #: Set to True if get_current_version should return a version if the command
    #: or package exists
    #: Set to False if get_current_version may not be able to return the
    #: version, even if the command or package is installed or present
    require_current_version: bool = True

    @property
    def value(
        self,
    ) -> t.Tuple[
        t.Union[str, None], t.Union[t.Callable, None], t.Union[t.Tuple[int], None]
    ]:
        """Get the package name, comparison operator and version tuple."""
        value = Check.value.fget(self)
        name, op, version = name_and_version(value)
        return name, op, version

    @value.setter
    def value(self, v):
        Check.value.fset(self, v)

    @abstractmethod
    def get_current_version(self) -> t.Union[None, t.Tuple[int]]:
        """Get the current version as a tuple of integers, or None if it can't
        be found."""
        return None

    def check(self, executor: t.Optional[Executor] = None, level: int = 0) -> Result:
        """Check whether the current version is compatible with the version
        specified in the value."""
        name, op, version = self.value
        current_version = self.get_current_version()

        if name is None:
            status = "failed (missing)"
        elif self.require_current_version and current_version is None:
            status = "failed (missing)"
        else:
            if version is not None and current_version is None:
                status = "present but current version unknown"
            elif (
                version is not None
                and current_version is not None
                and op is not None
                and not op(current_version, version)
            ):
                status = (
                    f"failed (incorrect version="
                    f"{'.'.join(map(str, current_version))})"
                )
            else:
                status = "passed"

        return Result(msg=self.msg.format(check=self), status=status)



"""Fixtures for project testing"""
from pathlib import Path

import pytest


@pytest.fixture
def test_env_file() -> dict:
    """Filename and expected values for the 'environment/test.env' file."""
    return {
        "filepath": str(Path(__file__).parent / "data" / "test.env"),
        "variables": {
            "VALUE1": "My Value",
            "VALUE2": "dev",
            "VALUE3": "my-dev",
            "VALUE4": "A Multiline\nenvironment variable",
            "VALUE5": "Extra endspaces removed",
        },
    }



"""Test the functionality of CheckAws"""
import pytest

from geomancy.checks.aws.base import CheckAws, CheckException


@pytest.mark.block_network
def test_check_aws_client():
    """Test the CheckAws base class client method"""
    # 1. Test with an invalid profile
    profile = "invalid_profile"
    check = CheckAws(name="CheckAws", profile=profile)
    assert check.profile == profile

    with pytest.raises(CheckException, match="profile not found"):
        check.client("s3")

    # 2. The default should return a valid client
    check = CheckAws(name="CheckAws")
    check.client("s3")


# noinspection GrazieInspection
@pytest.mark.block_network
def test_check_aws_username_invalid_profile():
    """Test the CheckAws.username() method with an invalid profile"""
    profile = "invalid_profile"
    check = CheckAws(name="CheckAws", profile=profile)

    with pytest.raises(CheckException, match="profile not found"):
        check.username()


# noinspection GrazieInspection
@pytest.mark.vcr
def test_check_aws_username_invalid_keys(aws_invalid):
    """Test the CheckAws.username() method with invalid AWS keys"""
    check = CheckAws(name="CheckAws")

    with pytest.raises(CheckException, match="could not connect to client"):
        check.username()


# noinspection GrazieInspection
@pytest.mark.vcr
def test_check_aws_username():
    """Test the CheckAws.username() method with valid credentials"""
    check = CheckAws(name="CheckAws")
    assert check.username() == "mytestuser"



"""Functions to load dotenv files and parse strings in dotenv format"""
import typing as t
import os
import re
import codecs
import logging
from pathlib import Path

__all__ = ("sub_env", "parse_env", "load_env")

logger = logging.getLogger(__name__)

#: Regex to match environment variables for subsitution--e.g. ${NAME} or $NAME
sub_re = re.compile(
    r"[$]"  # Start with a '$'. e.g. $NAME
    r"((?P<name_nobrace>[a-zA-Z_][a-zA-Z0-9_:\-?+]*)|"  # e.g. $NAME
    r" \{(?P<name_brace>[a-zA-Z_][a-zA-Z0-9_\s:\-?+]*)\})",  # e.g ${NAME}
    re.VERBOSE,
)

#: Regex to identify alternate variables from variable names
#: e.g. ${NAME:-default}
sub_alt_re = re.compile(
    # Does not require a brace. e.g. $NAME
    r"(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)"
    # May have an alternative descriptor
    r"((?P<default>:-|-)|(?P<error>:\?|\?)|(?P<replace>:\+|\+))?(?P<alt>[\w\s]*)",
    re.VERBOSE,
)

#: Regex to match environment variable names
env_name = r"(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)"  # env variable name
env_name_re = re.compile(env_name)

#: Regex to match environment variable values with substitution
env_value = (
    # Quoted value--e.g. "My $VAR" or 'My $VAR'--allowing for escaped quotes
    r"""((?P<quote>["|']{1,3})(?P<qvalue>(?:\\.|[^"'\\])+)(?P=quote)[^'"\n]*|"""
    # Unquoted value--e.g. My $VAR
    r"""(?P<value>[^'"\n]+))"""
)
env_value_re = re.compile(
    env_value,
    re.MULTILINE | re.DOTALL,
)

#: Regex to match "name=value" pairs from an env file
env_name_value_re = re.compile(
    r"^\s*{env_name}\s*=\s*{env_value}\s*$".format(
        env_name=env_name, env_value=env_value
    ),
    re.MULTILINE | re.DOTALL,
)

#: Regex to strip comments to the end of a line--# and not escaped values, \#
comment_re = re.compile(r"(^|\s+)(#.+)$")

#: Regex to strip backslashes from escaped quotes.
#: e.g. r"Let\'s go" -> r"Let's go"
escaped_quote_re = re.compile(r"\\(['\"])")


def sub_env(
    string: str, missing_default: str = "", strip_values: bool = True, **kwargs
) -> str:
    """Try to substitute environment variables in the string.

    Parameters
    ----------
    string
        The string to substituted
    missing_default
        Missing environment variables will have this value placed instead
    strip_values
        Remove whitespace at the start and end of non-quoted values
    kwargs
        In addition to os.environ, search the given kwargs for matches.

    Raises
    ------
    EnvironmentError
        Raised if an environment variable was not found and the ``:?``/``?``
        error directive was specified
        e.g. ``${MISSING?not found!}``

    Returns
    -------
    substituted_str
        The string with environment variables substituted

    Notes
    -----
    This function follows the docker compose_ format.

    - Environment variable names are preceded with a ``$`` character and may
      include braces. e.g. ``$VAR_NAME`` or ``${VAR_NAME}``
    - Environment variables names may include directives for default values
      (``${MISSING-default}``), errors for missing values (``${missing?error``)
      or replacement values (``${MISSING+replace}``)

    .. _compose: https://docs.docker.com/compose/environment-variables/env-file/
    """

    # Substitute environment variables in values
    def sub_func(m: re.Match):
        """Substitute a regex match from the environment variables, if possible,
        or return unmodified"""
        # Get the variable name, which may include alternates identified by
        # :-/-/:?/?/:+/"
        d = m.groupdict()
        name = d["name_brace"] if d["name_brace"] is not None else d["name_nobrace"]

        # Parse the alternate values
        alt_m = sub_alt_re.match(name)
        alt_d = alt_m.groupdict() if alt_m is not None else None

        name = alt_d["name"] if alt_d and alt_d["name"] else name
        default = alt_d["alt"] if alt_d and alt_d["default"] else None
        error = alt_d["alt"] if alt_d and alt_d["error"] else None
        replace = alt_d["alt"] if alt_d and alt_d["replace"] else None

        if name in os.environ:
            # found match in environment variables ('replace' will
            # replace the returned value)
            return os.environ[name] if replace is None else replace

        elif name in kwargs and not replace:
            # found match in passed keyword arguments (replace will replace its value)
            return kwargs[name] if replace is None else replace

        elif default is not None:  # Not found, return default if available
            return default

        elif error is not None:  # Not found, raise exception
            raise EnvironmentError(error)

        else:
            return missing_default

    # Parse the string like an environment variable value, which may contain
    # single quotes, double quotes or may be unquoted
    match = env_value_re.match(string)
    groupdict = match.groupdict() if match is not None else None
    if groupdict is None:
        return string

    # Try to parse the value based on the type of quoting
    if "value" in groupdict and groupdict["value"]:
        # Strip comments for non-quoted strings
        value = groupdict["value"]
        value = comment_re.sub(r"\1", value)  # Remove comments

        # Substitute values for non-quoted values
        value = sub_re.sub(sub_func, value)

        # Strip whitespace, if specified
        return value.strip() if strip_values else value

    elif "qvalue" in groupdict and groupdict["qvalue"]:
        # Retrieve quoted value and quote type
        value = groupdict["qvalue"]
        quote = groupdict["quote"]

        # Substitute escaped quotes
        value = escaped_quote_re.sub(r"\1", value)

        # Double-quote values may be substituted
        if '"' in quote:  # double quoted
            # process escape characters, e.g. \\t -> \t
            value = codecs.decode(value, "unicode_escape")

            # substitute values for double quoted values
            return sub_re.sub(sub_func, value)

        # Single-quoted values are used literally--i.e. without substitution
        elif "'" in quote:
            return value

    else:
        raise NotImplementedError


def parse_env(string: str, strip_values: bool = True) -> dict:
    """Parse a string in env format into a dict.

    See :func:`sub_env` for details on substitution.

    Parameters
    ----------
    string
        The string in env format to parse
    strip_values
        Remove whitespace at the start and end of non-quoted values

    Returns
    -------
    env_vars
        The parsed environment variables from the string. The variable names
        are dict keys and the variable values are dict values.
    """

    # Convert string into a dict
    env_vars = dict()
    for match in env_name_value_re.finditer(string):
        name = match.group(1)  # name (group 1)

        # env_value match. Keep full group, including quotes so that sub_env can parse
        # the value
        rest = match.group(2)

        # Substitute environment variables in the value
        value = sub_env(rest, strip_values=strip_values, **env_vars)

        # Add the new name-value pair in the env_vars
        env_vars[name] = value

    return env_vars


def load_env(
    filepath: t.Union[str, Path], overwrite: bool = False, *args, **kwargs
) -> dict:
    """Load an environment file.

    Parameters
    ----------
    filepath
        The path to the file with environment settings to load
    overwrite
        If True, overwrite environment variables that already exist
        If False (default), only load environment variables that don't already
        exist
    args, kwargs
        Arguments and keyword arguments passed to :func:`parse_env_str`

    Returns
    -------
    env
        A dict with all of the loaded env variable name-value pairs in a dict
    """
    # Try loading the file
    filepath = Path(filepath)
    try:
        string = filepath.read_text()
    except FileNotFoundError:
        logger.error(f"Could not file the file '{filepath}'")
        return 0

    # Parse the environment
    env_vars = parse_env(string=string, *args, **kwargs)

    # Load the environment variables
    updated_env_vars = dict()
    for name, value in env_vars.items():
        # Do nothing if the variable already exists, and overwrite isn't
        # specified
        if name in os.environ and not overwrite:
            continue

        os.environ[name] = value
        updated_env_vars[name] = value

        logger.debug(f"Substituted environment variable {name}={value}")

    return updated_env_vars



"""Tests for the CheckAWSIAM* classes"""
import datetime
from unittest.mock import patch

import pytest

from geomancy.checks.aws.base import CheckAws
from geomancy.checks.aws.iam import (
    CheckAwsIamAuthentication,
    CheckAwsIamRootAccess,
    CheckAwsIamAccessKeyAge,
    CheckAwsIam,
)


@pytest.mark.parametrize(
    "cls",
    (
        CheckAwsIam,
        CheckAwsIamAccessKeyAge,
        CheckAwsIamRootAccess,
        CheckAwsIamAuthentication,
    ),
)
def test_check_aws_iam_profile(cls):
    """Test that CheckAwsIam* profiles are properly loaded"""
    # Create the check with a custom profile
    custom_profile = "custom_profile"
    check = cls(name="CheckProfile", profile=custom_profile)

    assert check.profile == custom_profile

    # If the class has children (CheckAwsS3) check those too
    for child in check.children:
        if not isinstance(child, CheckAws):
            continue
        assert child.profile == custom_profile


@pytest.mark.vcr
def test_check_aws_iam_authentication():
    """Test the CheckAwsIamAuthentication check"""
    check = CheckAwsIamAuthentication(name="Checker")
    result = check.check()

    assert result.passed
    assert result.status.startswith("passed")


@pytest.mark.block_network
def test_check_aws_iam_authentication_invalid_profile():
    """Test the CheckAwsIamAuthentication check with and invalid profile"""
    # 1. Test with an invalid profile
    profile = "invalid_profile"
    check = CheckAwsIamAuthentication(name="Checker", profile=profile)
    assert check.profile == profile

    result = check.check()
    assert not result.passed
    assert result.status == "failed (profile not found)"


@pytest.mark.vcr
def test_check_aws_iam_authentication_invalid_keys(aws_invalid):
    """Test the CheckAwsIamAuthentication check with invalid AWS keys"""
    check = CheckAwsIamAuthentication(name="Checker")
    assert check.profile is None

    result = check.check()
    assert not result.passed
    assert result.status == "failed (can't authenticate with given tokens)"


@pytest.mark.parametrize("mode", ("has_root_keys", "has_signing_certs", "neither"))
@pytest.mark.vcr
def test_check_aws_iam_root_access(mode):
    """Test the CheckAwsIamRootAccess check"""
    check = CheckAwsIamRootAccess(name="CheckRoot")
    result = check.check()

    if mode == "has_root_keys":
        assert not result.passed
        assert result.status == "failed (account has root access keys)"
    elif mode == "has_signing_certs":
        assert not result.passed
        assert result.status == "failed (account has signing certificates)"
    else:
        assert result.passed


@pytest.mark.parametrize("age_days", (30, 120))
@pytest.mark.vcr
def test_check_aws_iam_access_key_age(age_days):
    """Test the CheckAwsIamAccessKeyAge check"""
    # The following test was run with and the cassette modified to meet
    # the number of days in 'age_days'
    check = CheckAwsIamAccessKeyAge(name="AccessKeyAge")

    # Create a mockdatetime to pin a date
    class mockdatetime(datetime.datetime):
        @classmethod
        def now(cls, *args, **kwargs):
            return datetime.datetime(2023, 8, 4, 5, 31, 11, 895114)

    # Run the check with a pinned date/time
    with patch("geomancy.checks.aws.iam.datetime.datetime", mockdatetime):
        result = check.check()

    # Check the age in days
    if age_days < 90:
        assert result.passed
    else:
        assert not result.passed


@pytest.mark.vcr
def test_check_aws_iam():
    """Test the CheckAwsIam check with a valid profile"""
    check = CheckAwsIam(name="CheckIAM")

    # By default, the other CheckAwsIam classes should be children
    for cls in (
        CheckAwsIamAccessKeyAge,
        CheckAwsIamRootAccess,
        CheckAwsIamAuthentication,
    ):
        assert cls in [type(child) for child in check.children]

    # The test should pass normally for a valid profile (see cassette)
    result = check.check()
    assert result.passed



"""
Test the environment variable check class
"""
import os

import pytest
from pytest import MonkeyPatch

from geomancy.checks import CheckEnv


def test_check_env_missing():
    """Test CheckEnv for a missing environment variable"""
    with MonkeyPatch.context() as mp:
        mp.delenv("MISSING", raising=False)
        check = CheckEnv(name="missing", value="${MISSING}")
        assert not check.check().passed


def test_check_env_present():
    """Test CheckEnv for a present environment variable"""
    with MonkeyPatch.context() as mp:
        mp.setenv("PRESENT", "VALUE")

        check = CheckEnv(name="PRESENT", value="${PRESENT}")
        assert check.name == "PRESENT"
        assert check.value == "VALUE"
        assert check.check().passed


def test_check_env_name_substitution(
    variable_name="VARIABLE_NAME", variable_value="PATH"
):
    """Test CheckEnv with substituting variable names with environment
    variables"""
    # Set the VariableName environment variable to be the name of another
    # environment variable that exists
    with MonkeyPatch.context() as mp:
        mp.setenv(variable_name, variable_value)

        # Setup the check to use the variable name with environment variable
        # substitution--i.e. wrap it in curly braces
        check = CheckEnv(name="substitution", value=f"${variable_name}")

        # Check that the subsitution is correct
        assert check.name == "substitution"
        assert check.raw_value == f"${variable_name}"  # unprocessed value
        assert check.value == variable_value

        assert check.check().passed


@pytest.mark.parametrize("regex", ("dev|prod|local", "(dev|prod|local)"))
def test_check_env_regex(regex):
    variable_name = "VARIABLE_NAME"
    variable_value = "dev"

    # Make sure the variable name isn't in the environment already
    assert variable_name not in os.environ

    # Place the variable in the environment and test the regex
    with MonkeyPatch.context() as mp:
        mp.setenv(variable_name, variable_value)

        # 1. The regex should match
        check1 = CheckEnv(name="regex1", value=f"${variable_name}", regex=regex)
        result = check1.check()
        assert result.passed
        assert result.status == "passed"

        # Change the variable value to something else in which the regex will
        # fail
        mp.setenv(variable_name, f"!{variable_value}!")

        # 2. The regex should not match
        check2 = CheckEnv(name="regex2", value=f"${variable_name}", regex=regex)
        result = check2.check()
        assert not result.passed
        assert result.status == f"failed (does not match regex '{regex}')"


def test_check_base_types_dict():
    """Test the CheckBase.types_dict() method for the CheckEnv class."""
    types = CheckEnv.types()

    # Verify CheckEnv entries and aliases
    assert types["CheckEnv"] == CheckEnv
    assert types["checkEnv"] == CheckEnv



"""Check classes to test running checks"""
import typing as t
import time

from .base import Check, Result, Executor


class CheckSleep(Check):
    """A check that sleeps for a specified about of time"""

    #: Time in seconds to sleep the check
    sleep: int = 3

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sleep = int(self.value)

    def check(self, executor: t.Optional[Executor] = None, level: int = 0) -> Result:
        time.sleep(self.sleep)
        return Result(status="passed", msg="" f"{self.name} sleeping for {self.sleep}s")



from . import checks, config, environment

# Project version
__version__ = (1, 1, 3)  # Major, minor, patch, stage


def get_version(version=__version__):
    """Generate a string from the __version__

    Examples
    --------
    >>> get_version((0, 1, 0, 'b'))
    '0.1.0b'
    >>> get_version((1, 0, 0))
    '1.0.0'
    """
    if isinstance(version[-1], str):
        return ".".join(map(str, version[:-1])) + version[-1]
    else:
        return ".".join(map(str, version))



"""Tests for CheckAWSS3"""
import pytest


from geomancy.checks.aws.base import CheckAws
from geomancy.checks.aws.s3 import (
    CheckAwsS3,
    CheckAwsS3BucketAccess,
    CheckAwsS3BucketPrivate,
)

# Basic accessibility tests


@pytest.mark.parametrize(
    "cls", (CheckAwsS3, CheckAwsS3BucketAccess, CheckAwsS3BucketPrivate)
)
def test_check_aws_s3_profile(cls):
    """Test that CheckAwsS3* profiles are properly loaded"""
    # Create the check with a custom profile
    custom_profile = "custom_profile"
    check = cls(name="CheckProfile", profile=custom_profile)

    assert check.profile == custom_profile

    # If the class has children (CheckAwsS3) check those too
    for child in check.children:
        if not isinstance(child, CheckAws):
            continue
        assert child.profile == custom_profile


@pytest.mark.vcr
def test_check_aws_s3_missing():
    """Test CheckAwsS3/CheckAwsS3BucketAccess check with a missing bucket"""
    bucket_name = "my-missing-bucket-noexist"

    # Run the check, which should not pass
    check = CheckAwsS3(name="Missing Bucket", value=bucket_name)
    result = check.check()
    assert not result.passed

    # Validate the result of the first child check, CheckAwsS3BucketAccess,
    # which could not find the bucket
    child_results = result.children
    assert not child_results[0].passed
    assert child_results[0].status == "failed (not found)"


@pytest.mark.vcr
def test_check_aws_s3_forbidden():
    """Test CheckAwsS3/CheckAwsS3BucketAccess check with a bucket that has
    forbidden access"""
    bucket_name = "docker"

    # Run the check, which should not pass
    check = CheckAwsS3(name="Forbidden Bucket", value=bucket_name)
    result = check.check()
    assert not result.passed

    # Validate the result of the first child check, CheckAwsS3BucketAccess,
    # which found that the bucket access is forbidden
    child_results = result.children
    assert not child_results[0].passed
    assert child_results[0].status == "failed (access forbidden)"


def test_check_aws_s3_invalid():
    """Test CheckAwsS3/CheckAwsS3BucketAccess check with a bucket that has an
    invalid name"""
    bucket_name = "!invalid!"

    # Run the check, which should not pass
    check = CheckAwsS3(name="Invalid Bucket", value=bucket_name)
    result = check.check()
    assert not result.passed

    # Validate the result of the first child check, CheckAwsS3BucketAccess,
    # which found the bucket name to be invalid
    child_results = result.children
    assert not child_results[0].passed
    assert child_results[0].status.startswith("failed (invalid bucket name")


# The following are public/private bucket tests
@pytest.mark.vcr
def test_check_aws_s3_without_public_access_block():
    """Test CheckAwsS3/CheckAwsS3BucketPrivate check using a bucket with
    Public Access Block turned off."""
    # The following is a bucket created with "Block Public Access" turned off
    # but without a policy or ACL specified. Altogether, it should not give
    # public access
    bucket_name = "testbucket-nopublicblock"

    # Run the check, which should not pass
    check = CheckAwsS3(name="Public Bucket", value=bucket_name)
    result = check.check()
    assert result.passed

    # Validate the result of the second child check, CheckAwsS3BucketAccess,
    # which found the bucket name to be invalid
    child_results = result.children
    assert child_results[1].passed


@pytest.mark.vcr
def test_check_aws_s3_public_policy():
    """Test CheckAwsS3/CheckAwsS3BucketPrivate check using a bucket with a
    public Policy."""
    # The following is a bucket created with "Block Public Access" turned off
    # and with a policy that allows public access. This bucket should be
    # publicly accessible and will fail the test
    bucket_name = "testbucket-public-policy"

    # Run the check, which should not pass
    check = CheckAwsS3(name="Public Bucket", value=bucket_name)
    result = check.check()
    assert not result.passed

    # Validate the result of the second child check, CheckAwsS3BucketAccess,
    # which found the bucket name to be invalid
    child_results = result.children
    assert not child_results[1].passed
    assert child_results[1].status == "failed (publicly accessible)"


@pytest.mark.vcr
def test_check_aws_s3_public_acl():
    """Test CheckAwsS3/CheckAwsS3BucketPrivate check using a bucket with a
    public ACL."""
    # The following is a bucket created with "Block Public Access" turned off
    # and with a policy that allows public access. This bucket should be
    # publicly accessible and will fail the test
    bucket_name = "testbucket-alluser-acl"

    # Run the check, which should not pass
    check = CheckAwsS3(name="Public Bucket", value=bucket_name)
    result = check.check()
    assert not result.passed

    # Validate the result of the second child check, CheckAwsS3BucketAccess,
    # which found the bucket name to be invalid
    child_results = result.children
    assert not child_results[1].passed
    assert child_results[1].status.startswith("failed (publicly accessible)")


@pytest.mark.vcr
def test_check_aws_s3_valid():
    """Test CheckAwsS3 check with a valid bucket"""
    bucket_name = "testbucket-valid"

    # Run the check, which should pass
    check = CheckAwsS3(name="Valid Bucket", value=bucket_name)
    result = check.check()
    assert result.passed

    # Validate the result of the first child check, CheckAwsS3BucketAccess,
    # which found an accessible bucket
    child_results = result.children
    assert child_results[0].passed
    assert child_results[0].status.startswith("passed")



"""Base class for AWS checks"""
import typing as t

from ..base import Check, Result, Executor, CheckException
from ...config import Parameter

__all__ = ("CheckAws",)

#: The error message to show when AWS modules can't be imported
import_error_msg = (
    "The 'aws' dependency is not installed: {exception}. "
    "Please reinstall with the '[aws]' or '[all]' extra install "
    '`python -m pip install "geomancy[aws]"` or '
    '`python -m pip install "geomancy[all]"`'
)


class CheckAws(Check):
    """Abstract base class for AWS checks"""

    import_error_msg = import_error_msg

    #: Profile name to use to authenticate the AWS client, str
    profile: t.Optional[str]

    #: Default for profile
    profile_default = Parameter("CHECK_AWS.PROFILE", None)

    def __init__(self, *args, profile: t.Optional[str] = None, **kwargs):
        self.profile = profile if profile is not None else self.profile_default
        super().__init__(*args, **kwargs)

    def client(self, *args, **kwargs) -> "botocore.client.BaseClient":
        """Retrieve the AWS client using the given profile.

        Raises
        ------
        CheckException
            The specified profile name could not be found
        """
        # Get the needed modules, profile name
        boto3, exceptions = self.import_modules("boto3", "botocore.exceptions")
        profile = kwargs.get("profile", self.profile)

        # Get a session
        try:
            session = (
                boto3.Session(profile_name=profile)
                if self.profile is not None
                else boto3.Session()
            )
        except exceptions.ProfileNotFound:
            raise CheckException("failed (profile not found)")

        return session.client(*args, **kwargs)

    def username(self, *args, **kwargs) -> str:
        """Retrieve the username of the current profile.

        Raises
        ------
        CheckException
            If the session profile was not found or the username could not be
            parsed
        """
        exceptions = self.import_modules("botocore.exceptions")

        iam = self.client("iam", *args, **kwargs)
        try:
            response = iam.get_user()
            return response["User"]["UserName"]
        except exceptions.ClientError:
            raise CheckException("failed (could not connect to client)")
        except KeyError:
            raise CheckException("failed (could not parse IAM.get_user())")

    def check(self, executor: t.Optional[Executor] = None, level: int = 0) -> Result:
        """Run sub-checks in sequence rather than concurrently, as done by the
        Check base class.

        This is done because the results of latter checks depend on earlier checks.
        """
        # check children
        child_results = []
        for child in self.children:
            result = child.check(executor=executor, level=level + 1)
            child_results.append(result)
            if not result.status.startswith("passed"):
                break

        return Result(
            msg=f"[bold]{self.name}[/bold]",
            children=child_results,
            condition=self.condition,
        )



"""The geo CLI entrypoint"""
import logging, logging.config
import os
from pathlib import Path

import click
from click_default_group import DefaultGroup

from .check import check_cmd
from .run import run_cmd
from .config import config_cmd
from .. import get_version

logger = logging.getLogger(__name__)

description = (Path(__file__).parent / ".." / "__description__.txt").read_text().strip()


def print_version(context, parameter, value):
    """Print the version"""
    if not value or context.resilient_parsing:
        return
    click.echo(get_version())
    context.exit()


@click.group(
    cls=DefaultGroup, default="check", default_if_no_args=True, help=description
)
@click.option("--debug", "-d", is_flag=True, help="Enable debugging information")
@click.option(
    "--version",
    "-V",
    is_flag=True,
    is_eager=True,
    expose_value=False,
    callback=print_version,
    help="Show the version and exit.",
)
@click.option("--disable-color", is_flag=True, help="Disable terminal coloring")
@click.pass_context
def geo_cli(ctx, debug, disable_color):
    """The main entrypoint for the 'geo' CLI"""
    # Disable coloring, if specified
    if disable_color:
        os.environ["NO_COLOR"] = "TRUE"

    # Setup logging
    logging.config.dictConfig(
        {
            "version": 1,
            "formatters": {
                "standard": {"format": "%(levelname)s:%(name)s: %(message)s"},
            },
            "handlers": {
                "default": {
                    "level": "DEBUG" if debug else "WARNING",
                    "formatter": "standard",
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stdout",  # Default is stderr
                },
            },
            "loggers": {
                "": {  # root logger
                    "handlers": ["default"],
                    "level": "WARNING",
                    "propagate": False,
                },
                "geomancy": {  # geomancy logs
                    "handlers": ["default"],
                    "level": "DEBUG" if debug else "INFO",
                    "propagate": False,
                },
            },
        }
    )


# Add sub-commands
geo_cli.add_command(check_cmd)  # noqa
geo_cli.add_command(run_cmd)  # noqa
geo_cli.add_command(config_cmd)  # noqa



"""
The 'run' subcommand
"""
import subprocess
import logging

import click

from .environment import env_options

__all__ = ("run_cmd",)

logger = logging.getLogger(__name__)


@click.command(name="run", context_settings={"ignore_unknown_options": True})
@env_options
@click.argument("args", nargs=-1)
def run_cmd(args, env):
    """Run command within environment"""
    logger.debug(f"args={args}, env={env}")

    # Run the command
    result = subprocess.run(args, env=env)
    return result.returncode



"""An instance-wide configuration"""
import typing as t
from threading import Lock
from pathlib import Path
from collections.abc import Mapping
import re
import tomllib
import yaml
import logging

__all__ = ("ConfigException", "Config", "Parameter")

logger = logging.getLogger(__name__)


class ConfigException(Exception):
    """An exception raised in validating or modifying the config"""


class ConfigMeta(type):
    """A metaclass to set up the creation of Config objects and singleton"""

    def __call__(cls, root: bool = True, **kwargs):
        # noinspection PyArgumentList
        obj = cls.__new__(cls, root=root)
        obj.__init__(**kwargs)
        return obj


class Config(metaclass=ConfigMeta):
    """A thread-safe Config manager to get and set configuration parameters.

    Notes
    -----
    The base class is deliberately modified dynamically with descriptors. This
    is because the Config was designed to be configured on the fly.
    """

    #: The root singleton instance
    _instance: t.Optional["Config"] = None

    #: The thread lock
    _lock: Lock = Lock()

    #: The regex to validate key names
    key_regex = re.compile(r"^[_A-Za-z][_A-Za-z0-9]*$")

    #: Different names for subsections in a dict which could contain config settings
    section_aliases = ("config", "Config")

    def __new__(cls, root: bool = True):
        # Create the singleton instance if it hasn't been created
        if cls._instance is None:
            # Lock the thread
            with cls._lock:
                # Prevent another thread from creating the instance in the
                # interim
                if cls._instance is None:
                    cls._instance = super().__new__(cls)

        if not root:
            # If this is not the root Config--i.e. a sub config--return a
            # new Config object instead of the root singleton
            obj = super().__new__(cls)
            return obj
        else:
            # Otherwise return the singleton
            return cls._instance

    def __init__(self, **kwargs):
        # Set the specified parameters
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __getattribute__(self, key):
        """Get attributes Config with support for attribute nesting"""
        Config.validate(key)
        try:
            return super().__getattribute__(key)
        except AttributeError:
            # Create a dict by default for a missing attribute
            # This allows nested attribute access
            self.__dict__[key] = Config(root=False)
            return super().__getattribute__(key)

    def __setattr__(self, key, value):
        """Set attributes in the Config"""
        Config.validate(key)
        super().__setattr__(key, value)

    def __eq__(self, other):
        """Test whether this Config is equal to another"""
        if not type(self) == type(other):
            return False
        return self.__dict__ == other.__dict__

    def __len__(self):
        """The number of items in this Config"""
        return len(self.__dict__)

    @classmethod
    def validate(cls, key):
        """Ensure that the given key is valid"""
        if not cls.key_regex.match(key):
            raise ConfigException(
                f"The key '{key}' is not valid. "
                f"Must match the pattern: '{cls.key_regex.pattern}'"
            )

    def update(self, updates: Mapping) -> None:
        """Recursively update config with values from a dict.

        Parameters
        ----------
        updates
            The dict with values to recursively update

        Raises
        ------
        ConfigException
            Raised when the updates include a change that would replace a
            sub-config (subsection) that includes parameters with a single
            parameter.

        Notes
        -----
        - This function operates recursively to update sub-config (subsection)
          objects
        - This function does not allow a sub-config (subsection) to be replaced
          or truncated by a parameter. This will raise a ConfigException.
        - This function is not atomic. If a ConfigException is raised partway
          through the update, this config will only be partially updated.
        """
        for k, v in updates.items():
            current_value = getattr(self, k)  # get the current value or sub_config

            if isinstance(v, Mapping):
                # Use the corresponding update function. ex: dict
                current_value.update(v)
            elif isinstance(v, Config):
                current_value.update(v.__dict__)
            elif isinstance(current_value, Config) and len(current_value.__dict__) > 0:
                # In this case, the update value 'v' is a simple parameter but the
                # current_value is a sub-config with items in it. This is not allowed.
                raise ConfigException(
                    f"Cannot replace config section '{k}' with a parameter '{v}'"
                )
            else:
                logger.debug(f"Config.update(): set {k}={v}")
                setattr(self, k, v)

    @classmethod
    def load(cls, d: dict, root=True) -> "Config":
        """Load config with values from a dict.

        Parameters
        ----------
        d
            The dict with config values to load
        root
            Whether the root Config singleton is returned or a new sub config
            is returned.

        Returns
        -------
        config
            The root config instance with parameters loaded
        """
        config = cls(root=root)

        for k, v in d.items():
            # Only key strings are allowed for the Config
            assert isinstance(k, str), (
                f"Config keys must be strings. " f"Received: '{k}'"
            )

            if isinstance(v, dict):
                # Create a sub Config and load the sub dict
                sub_config = Config.load(v, root=False)
                setattr(config, k, sub_config)
            else:
                # Otherwise just store the value
                logger.debug(f"Config.load(): set {k}={v}")
                setattr(config, k, v)

        return config

    @classmethod
    def load_toml(cls, filename: Path) -> "Config":
        """Load config from a file formatted in TOML format.

        Returns
        -------
        config
            The root config instance with parameters loaded
        """
        with open(filename, "rb") as f:
            d = tomllib.load(f)

        # If the config dict is an item in 'd', get that instead
        for alt_name in cls.section_aliases:
            d = d[alt_name] if alt_name in d else d
        return cls.load(d)

    @classmethod
    def loads_toml(cls, s: str) -> "Config":
        """Load config from a string formatted in TOML format.

        Returns
        -------
        config
            The root config instance with parameters loaded
        """
        d = tomllib.loads(s)

        # If the config dict is an item in 'd', get that instead
        for alt_name in cls.section_aliases:
            d = d[alt_name] if alt_name in d else d
        return cls.load(d)

    def dumps_toml(self, name: str = "config", level: int = 0) -> str:
        """Produce a string of the current config in TOML format.

        Parameters
        ----------
        name
            Name of the current section. Subsections are identified with '.'
            characters
        level
            Current level of the hierarchy
        """
        indent = " " * 2 * level  # indentation
        text = f"{indent}[{name}]\n"  # section heading

        # Process general parameters before sections (sub configs)
        items = [
            (k, v)
            for k, v in sorted(self.__dict__.items())
            if not isinstance(v, Config)
        ]
        for key, value in items:
            text += indent  # Add indentation

            # Different simple parameter formats for TOML
            if isinstance(value, str) and "'" not in value:
                text += f"{key}='{value}'\n"
            elif isinstance(value, str):
                text += f'{key}="{value}"\n'
            elif isinstance(value, bool):
                text += f"{key}={str(value).lower()}\n"
            else:
                text += f"{key}={value}\n"

        if len(items) > 0:
            text += "\n"

        # Next process sections (sub configs)
        items = [
            (k, v) for k, v in sorted(self.__dict__.items()) if isinstance(v, Config)
        ]
        for key, value in items:
            text += indent  # Add indentation
            text += value.dumps_toml(name=".".join((name, key)), level=level + 1)

        if level == 0:
            text = text.rstrip("\n") + "\n"  # Remove multiple terminal newlines
        return text

    @classmethod
    def load_yaml(cls, filename: Path) -> "Config":
        """Load config from a file formatted in YAML format.

        Returns
        -------
        config
            The root config instance with parameters loaded
        """
        with open(filename, "r") as f:
            d = yaml.load(f, yaml.SafeLoader)

        # If the config dict is an item in 'd', get that instead
        for alt_name in cls.section_aliases:
            d = d[alt_name] if alt_name in d else d
        return cls.load(d)

    @classmethod
    def loads_yaml(cls, s: str) -> "Config":
        """Load config from a string formatted in YAML format.

        Returns
        -------
        config
            The root config instance with parameters loaded
        """
        d = yaml.load(s, yaml.SafeLoader)

        # If the config dict is an item in 'd', get that instead
        for alt_name in cls.section_aliases:
            d = d[alt_name] if alt_name in d else d
        return cls.load(d)

    def dumps_yaml(
        self, name: str = "config", indent_spaces: int = 2, level: int = 0
    ) -> str:
        """Produce a string of the current config in YAML format.

        Parameters
        ----------
        name
            Name of the current section.
        indent_spaces
            Number of spaces to indent subsections.
        level
            Current level of the hierarchy
        """
        indent = " " * indent_spaces * level  # indentation
        text = f"{indent}{name}:\n"  # section heading

        # Process general parameters before sections (sub configs)
        items = [
            (k, v)
            for k, v in sorted(self.__dict__.items())
            if not isinstance(v, Config)
        ]
        for key, value in items:
            text += indent + " " * indent_spaces  # Add indentation

            # Different simple parameter formats for TOML
            if isinstance(value, str) and "'" not in value:
                text += f"{key}: '{value}'\n"
            elif isinstance(value, str):
                text += f'{key}: "{value}"\n'
            elif isinstance(value, bool):
                text += f"{key}: {value}\n"
            else:
                text += f"{key}: {value}\n"

        if len(items) > 0:
            text += "\n"

        # Next process sections (sub configs)
        items = [
            (k, v) for k, v in sorted(self.__dict__.items()) if isinstance(v, Config)
        ]
        for key, value in items:
            text += indent  # Add indentation
            text += value.dumps_yaml(name=key, level=level + 1)

        if level == 0:
            text = text.rstrip("\n") + "\n"  # Remove multiple terminal newlines
        return text

    def pprint_toml(self):
        """Pretty print in TOML format."""
        print(self.dumps_toml())

    def pprint_yaml(self):
        """Pretty print in YAML format"""
        print(self.dumps_yaml())


# dummy placeholder object
missing = object()


class Parameter:
    """A descriptor for a Config parameter.

    Notes
    -----
    This is a descriptor. Its value can be accessed or modified from an
    instance. However, its value can only be access from the class itself. This
    is because the descriptor will be replaced if its corresponding class
    attribute is modified with a new value. Stated another way, the __set__
    method is not called from a class--only instances of a class.
    """

    __slots__ = ("key", "_config")

    #: The key/name of the parameter in the Config
    key: str

    #: The delimiter used for splitting keys
    delim: str = "."

    #: A reference to the Config() singleton
    _config: Config

    def __init__(self, key, default=missing):
        self.key = key
        self._config = Config()

        # Set the default, if specified
        if default is not missing:
            self.__set__(instance=None, value=default)

    def __repr__(self):
        value = getattr(self._config, self.key)
        return f"Param({self.key}={value})"

    def __get__(self, instance, objtype=None):
        # Convert strings with '.' into nested keys
        keys = self.key.split(self.delim)
        value = self._config

        for key in keys:
            value = getattr(value, key)
        return value

    def __set__(self, instance, value):
        # Convert strings with '.' into nested keys
        keys = self.key.split(self.delim)
        obj = self._config
        for key in keys[:-1]:
            obj = getattr(obj, key)
        setattr(obj, keys[-1], value)



"""
The 'config' subcommand
"""
import logging

import click

from ..config import Config

__all__ = ("config_cmd",)

logger = logging.getLogger(__name__)


@click.command(name="config")
@click.option("--toml", is_flag=True, help="Print default config in toml format")
@click.option("--yaml", is_flag=True, help="Print default config in yaml format")
def config_cmd(toml, yaml, **kwargs):
    """Configuration information"""
    logger.debug(f"toml={toml}, yaml={yaml}")

    conf = Config()
    if toml:
        conf.pprint_toml()
    else:
        conf.pprint_yaml()



"""Check AWS credentials, Identity and Access Management (`IAM`_) and security
best practices.

Checks include:

- Authentication with the default profile or a specified profile
- Access keys need to be rotated (age >90 days)
- Root keys and signing certificates have not been created.

.. _IAM: https://aws.amazon.com/iam/
"""
import typing as t
import logging
import datetime

from .base import CheckAws
from ..base import Result, Executor, CheckException
from ..utils import pop_first
from ...config import Parameter

logger = logging.getLogger(__name__)


class CheckAwsIamAuthentication(CheckAws):
    """Checks that the AWS profile can be authenticated"""

    msg = Parameter(
        "CHECK_AWS_IAM_AUTHENTICATION.MSG",
        default="Check AWS IAM authentication",
    )

    def check(self, executor: t.Optional[Executor] = None, level: int = 0) -> Result:
        msg = self.msg.format(check=self)
        exceptions = self.import_modules("botocore.exceptions")

        # Get the STS client to get details on the
        try:
            sts = self.client("sts")
        except CheckException as exc:
            return Result(status=exc.args[0], msg=msg)

        # Retrieve caller identity
        try:
            response = sts.get_caller_identity()
        except exceptions.ClientError:
            return Result(
                status="failed (can't authenticate with given tokens)", msg=msg
            )

        if "Arn" not in response:
            return Result(status="failed (can't verify identity with STS)", msg=msg)

        arn = response["Arn"]
        return Result(status=f"passed (ARN='{arn}')", msg=msg)


class CheckAwsIamAccessKeyAge(CheckAws):
    """Check the age of AWS access keys"""

    #: The key age in days
    key_age: int

    #: The default key age (in days)
    key_age_default: int = Parameter("CHECK_AWS_IAM_ACCESS_KEY_AGE.KEY_AGE", default=90)

    #: Aliases for the key_age parameter
    key_age_aliases = ("key_age", "age")

    msg = Parameter(
        "CHECK_AWS_IAM_ACCESS_KEY_AGE.MSG",
        default="Check AWS IAM access key age ({check.key_age} days)",
    )

    def __init__(self, *args, **kwargs):
        # Retrieve kwargs
        self.key_age = pop_first(
            kwargs, *self.key_age_aliases, default=self.key_age_default
        )
        super().__init__(*args, **kwargs)

    def check(self, executor: t.Optional[Executor] = None, level: int = 0) -> Result:
        msg = self.msg.format(check=self)
        exceptions = self.import_modules("botocore.exceptions")

        # Get the IAM client and username
        try:
            iam = self.client("iam")
            username = self.username()
        except CheckException as exc:
            return Result(status=exc.args[0], msg=msg)

        # Retrieve the list of access keys for the given username
        try:
            response = iam.list_access_keys(UserName=username)
            access_keys = response["AccessKeyMetadata"]
            access_key_dates = {d["AccessKeyId"]: d["CreateDate"] for d in access_keys}
        except KeyError:
            return Result(status="failed (could not parse list_access_key)")

        # Find which keys are too old
        old_keys = dict()
        for key_id, key_date in access_key_dates.items():
            delta = datetime.datetime.now(tz=None) - key_date.replace(tzinfo=None)

            if delta.days > self.key_age:
                old_keys[key_id] = key_date

        if old_keys:
            # There were keys older than self.ket_age, this is a fail
            return Result(
                status=(
                    f"failed ({' and '.join(old_keys)} are older than "
                    f"{self.key_age} days)"
                ),
                msg=msg,
            )
        else:
            return Result(status="passed", msg=msg)


class CheckAwsIamRootAccess(CheckAws):
    """Check the presence of AWS account (root) access Keys

    see: https://aws.amazon.com/blogs/security/an-easier-way-to-determine-the-presence-of-aws-account-access-keys/ # noqa
    """

    msg = Parameter(
        "CHECK_AWS_IAM_ROOT_ACCESS.MSG",
        default="Check AWS IAM root keys are not present",
    )

    def check(self, executor: t.Optional[Executor] = None, level: int = 0) -> Result:
        msg = self.msg.format(check=self)
        exceptions = self.import_modules("botocore.exceptions")

        # Get the IAM client
        try:
            iam = self.client("iam")
        except CheckException as exc:
            return Result(status=exc.args[0], msg=msg)

        # Get the account summary
        try:
            response = iam.get_account_summary()
        except exceptions.ClientError:
            return Result(status="failed (couldn't authenticate host)", msg=msg)

        # See if the expected items are in the response and fail if they aren't
        if (
            "SummaryMap" not in response
            or "AccountAccessKeysPresent" not in response["SummaryMap"]
            or "AccountSigningCertificatesPresent" not in response["SummaryMap"]
        ):
            logger.debug(str(response))
            return Result(status="failed (couldn't retrieve account summary)", msg=msg)

        summary = response["SummaryMap"]
        has_account_access_key = summary["AccountAccessKeysPresent"] != 0
        has_account_signing_certs = summary["AccountSigningCertificatesPresent"] != 0

        # Track statuses that amount to a fail
        statuses = []
        if has_account_access_key:
            statuses.append("account has root access keys")
        if has_account_signing_certs:
            statuses.append("account has signing certificates")

        if statuses:
            return Result(status=f"failed ({' and '.join(statuses)})", msg=msg)
        else:
            return Result(status="passed", msg=msg)


class CheckAwsIam(CheckAws):
    """Check the IAM access credentials and settings"""

    #: Check whether root secret keys or signing certs exist
    root_access: bool

    #: Default value for the root_access parameter
    root_access_default = Parameter("CHECK_AWS_IAM.ROOT_ACCESS", True)

    #: Aliases for the root_access parameter
    root_access_aliases = ("root_access", "root")

    #: Check the key age (in days)
    key_age: t.Optional[int]

    #: Default value for key_age (see :class:`CheckAwsIamAccessKeyAge`)
    key_age_default = Parameter("CHECK_AWS_IAM_ACCESS_KEY_AGE.KEY_AGE")

    #: Aliases for the key_age parameter
    key_age_aliases = ("key_age", "age")

    msg = Parameter(
        "CHECK_AWS_IAM.MSG",
        default="Check AWS IAM permissions",
    )

    aliases = ("checkAWSIAM", "checkAwsIAM", "CheckAwsIAM", "checkIAM", "CheckIAM")

    def __init__(self, *args, **kwargs):
        # Set up keyword arguments
        self.root_access = pop_first(
            kwargs, *self.root_access_aliases, default=self.root_access_default
        )
        self.key_age = pop_first(
            kwargs, *self.key_age_aliases, default=self.key_age_default
        )

        # Set up the rest of the class
        super().__init__(*args, **kwargs)

        # Replace children
        self.children.clear()

        # Add sub-checks
        # 1. CheckAwsIamAuthentication
        child = CheckAwsIamAuthentication(*args, **kwargs)
        child.name = f"{self.name}Authentication"
        self.children.append(child)

        # 2. CheckAwsIamAccessKeyAge
        if isinstance(self.key_age, int):
            child = CheckAwsIamAccessKeyAge(*args, **kwargs)
            child.name = f"{self.name}AccessKeyAge"
            self.children.append(child)

        # 3. CheckAwsIAMRootAccess
        if self.root_access:
            child = CheckAwsIamRootAccess(*args, **kwargs)
            child.name = f"{self.name}RootAccess"
            self.children.append(child)



"""
Check the existence and, optionally, the value of an environment variable.
"""
import typing as t
import re

from .base import Check, Result, Executor
from ..environment import sub_env
from ..config import Parameter


class CheckEnv(Check):
    """Check the current environment variables."""

    #: (Optional) regex to match the environment variable value
    regex: t.Optional[t.Tuple[str, ...]] = None

    msg = Parameter(
        "CHECK_ENV.MSG",
        default="Check environment variable '{check.raw_value}'",
    )

    aliases = ("checkEnv",)

    def __init__(self, *args, regex: t.Optional[str] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.regex = regex

    def check(self, executor: t.Optional[Executor] = None, level: int = 0) -> Result:
        """Check the environment variable value."""
        # Substitute environment variables, if needed
        value = sub_env(self.raw_value) if self.raw_value is not None else None

        if value is None:
            # If the value is None, the environment variable doesn't exist.
            status = "failed (missing)"
        elif value == "":
            # An empty string environment variable is considered not set
            status = "failed (empty string)"
        elif isinstance(self.regex, str) and re.match(self.regex, value) is None:
            # Check the regex, if specified
            status = f"failed (does not match regex '{self.regex}')"
        else:
            # All checks passed!
            status = "passed"

        msg = self.msg.format(check=self, status=status)
        return Result(msg=msg, status=status)



from .base import Check, CheckException, Result
from .env import CheckEnv
from .exec import CheckExec
from .path import CheckPath
from .platform import CheckPlatform
from .python import CheckPythonPackage
from .test import CheckSleep

from . import aws

__all__ = (
    Check,
    CheckException,
    Result,
    CheckEnv,
    CheckExec,
    CheckPath,
    CheckPlatform,
    CheckPythonPackage,
    CheckSleep,
    aws,
)



"""Fixtures for AWS tests"""
import re

import pytest


test_aws_username = "mytestuser"
test_aws_account_id = "8" * 12
test_aws_user_id = "X" * 20
test_aws_access_key = "Y" * 20
test_aws_secret_key = "S" * 40
test_aws_request_id = "99999999-9999-9999-9999-999999999999"


@pytest.fixture(scope="module")
def vcr_config():
    """VCR additional configuration for AWS"""

    # Scrub functions
    def scrub_request(request):
        """Script AWS PII from requence"""
        # Strip content-type headers
        request.headers.clear()

        # Replace body string
        if request.body is not None:
            body = request.body.decode("UTF-8")

            # username substitutions
            body = re.sub(r"(?<=UserName=)(\w+)", test_aws_username, body)

            request.body = body.encode("ascii")
        return request

    def scrub_response(response):
        """Scrub AWS PII from response"""
        # Remove unneeded header entries
        for entry in ("x-amz-request-id", "x-amz-id-2"):
            response["headers"].pop(entry, None)

        # Substitute strings in body
        body = response["body"]
        s = body["string"].decode("UTF-8")  # bytes -> string

        # username substitutions
        s = re.sub(r"(?<=user/)(\w+)", test_aws_username, s)
        s = re.sub(r"(?<=<UserName>)(\w+)", test_aws_username, s)

        # Account ID substitutions
        s = re.sub(r"\d{12}", test_aws_account_id, s)

        # User ID substitutions
        s = re.sub(r"AIDA[A-Z0-9]{1,124}", test_aws_account_id, s)  # User Id

        # Access Key ID substitutions
        s = re.sub(r"(?<![A-Z0-9])[A-Z0-9]{20}(?![A-Z0-9])", test_aws_access_key, s)

        # Secret Key ID substitutions
        s = re.sub(
            r"(?<![A-Za-z0-9/+=])[A-Za-z0-9/+=]{40}(?![A-Za-z0-9/+=])",
            test_aws_secret_key,
            s,
        )

        # Request substitutions
        s = re.sub(
            r"(?<=<RequestId>)([a-f0-9\-]+)", test_aws_request_id, s
        )  # Request Id

        body["string"] = s.encode("ascii")  # string -> bytes

        # Replace or remove strings in headers
        response["headers"]["x-amzn-RequestId"] = test_aws_request_id
        return response

    vcr_config = dict()
    vcr_config["before_record_request"] = scrub_request
    vcr_config["before_record_response"] = scrub_response
    vcr_config["filter_headers"] = [
        "authorization",
        "User-Agent",
        "X-Amz-Date",
        "X-Amz-Content-SHA256",
        "amz-sdk-invocation-id",
        "amz-sdk-request",
    ]
    return vcr_config


@pytest.fixture
def aws_invalid(monkeypatch):
    """Take a Monkeypatch context and load invalid AWS credentials"""
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "INVALID")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "INVALID")
    monkeypatch.setenv("AWS_SESSION_TOKEN", "INVALID")



"""
Options and settings for environment files
"""
import typing as t
import logging

import click

from .utils import filepaths
from ..environment import load_env

__all__ = ("env_options",)

logger = logging.getLogger(__name__)


class RequiredOther(click.Option):
    """Make the use of option require the specification/presence of another
    option"""

    def __init__(self, *args, **kwargs):
        requires = kwargs.pop("requires")
        self.requires = [requires] if isinstance(requires, str) else requires

        assert self.requires, "'requires' parameter required"
        kwargs["help"] = (
            kwargs.get("help", "")
            + f" NOTE: This argument requires '{', '.join(self.requires)}' "
            f"options to be specified."
        ).strip()
        super().__init__(*args, **kwargs)

    def handle_parse_result(
        self, ctx: click.Context, opts: t.Mapping[str, t.Any], args: t.List[str]
    ) -> t.Tuple[t.Any, t.List[str]]:
        """Check to make sure that all other required options are specified if this
        option is specified."""
        # If this option is specified, check that the 'requires' field is
        # specified too
        if self.name in opts and not all(req in opts for req in self.requires):
            # Find the missing required parameters
            missing = [req for req in self.requires if req not in opts]

            # Raise an exception
            raise click.BadOptionUsage(
                self.name,
                f"The '{self.name}' option requires the following option(s): "
                f"{', '.join(missing)}",
                ctx=ctx,
            )

        return super().handle_parse_result(ctx, opts, args)


class EnvOption(click.Option):
    """Generate an environment variable dict from environment files"""

    # Whether to overwrite new values over existing values in the environment
    # (i.e. os.environ dict).
    overwrite = False

    def handle_parse_result(
        self, ctx: click.Context, opts: t.Mapping[str, t.Any], args: t.List[str]
    ) -> t.Tuple[t.Any, t.List[str]]:
        """Set the overwrite attribute if the --overwrite flag was set"""
        # Get options
        if "overwrite" in opts:
            self.overwrite = opts["overwrite"]
            logger.debug(f"Environment overwrite set to: {self.overwrite}")

        # Return as normal
        return super().handle_parse_result(ctx, opts, args)

    def process_value(self, ctx: click.Context, value: t.Any) -> t.Any:
        """Process the environment files (-e) option strings into an env dict.

        Returns
        -------
        env
            The processed env dict
        """
        # Set a default, if it's not specified
        value = tuple() if value is None else value

        # Retrieve the env_files from the arguments
        existing_paths = []
        for path in value:
            existing_paths += filepaths(path)

        # Load the environment files and keep track of the number of variables
        # substituted
        env = dict()
        for filepath in existing_paths:
            returned_dict = load_env(filepath, overwrite=self.overwrite)
            env.update(returned_dict)
        return env


# An option group that returns an env dict
def env_options(func=None):
    """Options for loading and using environment files"""
    opts = [
        click.option(
            "--overwrite",
            is_flag=True,
            help="Overwrite environment variables from environment file values",
            cls=RequiredOther,  # Requires the -e/--env options to be set
            requires="env",
            expose_value=False,  # Do not pass to inner command functions
        ),
        click.option(
            "--env",
            "-e",
            multiple=True,
            type=click.Path(exists=True),
            cls=EnvOption,
            help="Environment files to load",
        ),
    ]

    def wrap(inner_func):
        for opt in opts:
            inner_func = opt(inner_func)
        return inner_func

    wrap.__doc__ = env_options.__doc__

    return wrap if func is None else wrap(func)



"""
The 'check' subcommand
"""
import typing as t
import logging
import tomllib
import time
from concurrent.futures import ThreadPoolExecutor
from contextlib import ExitStack

import click
import yaml
from rich.live import Live
from rich.console import Group
from rich.rule import Rule
from rich.console import Console, Theme
import rich.progress as progress

from .environment import env_options
from .utils import filepaths
from ..checks import Check
from ..config import Config

__all__ = ("check_cmd",)

# Setup logger and configuration
logger = logging.getLogger(__name__)


# Exception classes
class MissingChecks(click.ClickException):
    """No checks were found in the checks files"""


# Config options
config = Config()
config.CLI.CHECKS_PATHS = [  # Default paths for checks files
    "pyproject.toml",
    ".geomancy.??ml",
    "geomancy.??ml",
    "geomancy.yml",
    ".geomancy.yml",
]
config.CLI.TOML_EXTS = [".toml"]  # Default file extensions for TOML files
config.CLI.YAML_EXTS = [".yml", ".yaml"]  # Default file extensions for YAML files


def validate_checks_files(
    ctx: click.Context, param: click.Parameter, values: t.Tuple[str]
):
    """Validate the checks files arguments and convert to valid paths"""
    # Convert filepath strings into Path objects. Use default locations if
    # no checks_files were specified (i.e. it is an empty list)
    existing_files = []
    for path in values or config.CLI.CHECKS_PATHS:
        existing_files += filepaths(path)

    # Nothing to do if no checks files were found
    if len(existing_files) == 0:
        raise click.MissingParameter(
            "Could not find a checks file.", ctx=ctx, param=param
        )
    logging.debug(f"Checking the following files: {existing_files}")
    return existing_files


# Setup 'check' command
@click.command(name="check")
@env_options
@click.argument("checks_files", nargs=-1, type=str, callback=validate_checks_files)
def check_cmd(checks_files, env):
    """Run checks"""
    logger.debug(f"check_files={checks_files}, env={env}")

    # Convert the checks_files into checks
    checks = []
    for checks_file in checks_files:
        # Parse the file by filetype
        if checks_file.suffix in config.CLI.TOML_EXTS:
            with open(checks_file, "rb") as f:
                d = tomllib.load(f)

        elif checks_file.suffix in config.CLI.YAML_EXTS:
            with open(checks_file, "r") as f:
                d = yaml.load(f, Loader=yaml.SafeLoader)

        else:
            continue

        # pyproject.toml files have their items placed under the [tool.geomancy]
        # section
        if checks_file.name == "pyproject.toml":
            d = d.get("tool", dict()).get("geomancy", dict())

        # Load config section, if available
        for config_name in config.section_aliases:
            config_section = d.pop(config_name, None)
            if isinstance(config_section, dict):
                config.update(config_section)

        # Load the rest into a root CheckBase
        check = Check.load(d, name=str(checks_file))
        if check is not None:
            checks.append(check)

    # Create a root check, if there are a lot of checks
    if len(checks) > 1:
        check = Check(name=f"Checking {len(checks)} files", children=checks)
    elif len(checks) == 1:
        check = checks[0]
    else:
        plural = True if len(checks_files) > 1 else False
        raise MissingChecks(
            f"No checks were found in the file{'s' if plural else ''}: "
            f"{', '.join(map(str, checks_files))}."
        )

    # Set up a console for rendering to the terminal
    console = Console(theme=Theme({"repr.number": ""}))

    with ExitStack() as stack:
        # Context manager for running checks in multiple threads
        executor = stack.enter_context(ThreadPoolExecutor())

        # Context manager for rendering live to the terminal (rich)
        live = stack.enter_context(Live(refresh_per_second=4, console=console))

        # Run the checks, display the results to the terminal
        result = check.check(executor=executor)

        # Get the total number of checks
        check_total = len(check.flatten)

        # Set up a progress bar and render group
        pbar = progress.Progress(
            progress.SpinnerColumn(),
            progress.BarColumn(),
            progress.TextColumn("checks"),
            progress.MofNCompleteColumn(),
            progress.TimeRemainingColumn(),
        )
        task1 = pbar.add_task("checking...", total=check_total)

        # Update the display until the checks are done
        while not result.done:
            time.sleep(0.5)

            # Update progress
            pbar.update(task1, completed=len(result.finished))

            # Update group
            group = Group(
                result.rich_table(),
                pbar.make_tasks_table(tasks=pbar.tasks),
            )
            live.update(group)

        # Create a summary line to render
        passed_total = len([r for r in result.finished if r.passed])
        failed_total = check_total - passed_total
        elapsed = sum(task.elapsed for task in pbar.tasks if task.elapsed is not None)

        if result.passed:
            color = "green"
            title = f"[{color}][bold]{passed_total} passed[/bold][/{color}]"
        else:
            color = "red"
            title = f"[{color}][bold]{failed_total} failed[/bold][/{color}]"

        title = f"{title}[{color}] in {elapsed:.2f}s[/{color}]"
        status = Rule(title=title, characters="=", style=color)

        # Print the final table and summary
        group = Group(result.rich_table(), status)
        live.update(group)

    if not result.passed:
        exit(1)

    return result.passed
